/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type Topology struct {

	Memory *Memory2 `json:"Memory,omitempty"`

	Processor *Processor2 `json:"Processor,omitempty"`
}
