/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

//  Storage runtime statistics
type StorageStats struct {

	ReadCountNormalized int32 `json:"ReadCountNormalized,omitempty"`

	ReadSizeBytes int32 `json:"ReadSizeBytes,omitempty"`

	WriteCountNormalized int32 `json:"WriteCountNormalized,omitempty"`

	WriteSizeBytes int32 `json:"WriteSizeBytes,omitempty"`
}
