/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type Networking struct {

	AllowUnqualifiedDnsQuery bool `json:"AllowUnqualifiedDnsQuery,omitempty"`

	DnsSearchList string `json:"DnsSearchList,omitempty"`

	NetworkSharedContainerName string `json:"NetworkSharedContainerName,omitempty"`

	//  Guid in windows; string in linux
	Namespace string `json:"Namespace,omitempty"`

	NetworkAdapters []string `json:"NetworkAdapters,omitempty"`
}
