package lcow

const (
	// DefaultScratchSizeGB is the size of the default LCOW scratch disk in GB
	DefaultScratchSizeGB = 20

	// defaultVhdxBlockSizeMB is the block-size for the scratch VHDx's this package can create.
	defaultVhdxBlockSizeMB = 1
)
